/*
 * Copyright 2012 JOHK Software
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.wesleyhome.johksoftware.utilities.dependency;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;

/**
 * @author Justin Wesley
 * @since 1.0
 *
 */
public class DependencySet<E extends Dependency<? extends DependencyIdentifier>> implements SortedSet<E> {

	private Map<DependencyIdentifier, E> dependancyMap;
	private List<DependencyIdentifier> visited;
	private List<E> sorted;
	private boolean isSorted = false;

	/**
	 * 
	 */
	public DependencySet() {
		dependancyMap = new LinkedHashMap<DependencyIdentifier, E>();
		visited = new ArrayList<DependencyIdentifier>();
		sorted = new ArrayList<E>();
	}

	public DependencySet(final Collection<E> col) {
		this();
		addAll(col);
	}

	/* (non-Javadoc)
	 * @see java.util.List#size()
	 */
	@Override
	public int size() {
		return getSortedList().size();
	}

	/* (non-Javadoc)
	 * @see java.util.List#isEmpty()
	 */
	@Override
	public boolean isEmpty() {
		return getSortedList().isEmpty();
	}

	/* (non-Javadoc)
	 * @see java.util.List#contains(java.lang.Object)
	 */
	@Override
	public boolean contains(final Object o) {
		return getSortedList().contains(o);
	}

	/* (non-Javadoc)
	 * @see java.util.List#iterator()
	 */
	@Override
	public Iterator<E> iterator() {
		return getSortedList().iterator();
	}

	/* (non-Javadoc)
	 * @see java.util.List#toArray()
	 */
	@Override
	public Object[] toArray() {
		return getSortedList().toArray();
	}

	/* (non-Javadoc)
	 * @see java.util.List#toArray(T[])
	 */
	@Override
	public <T> T[] toArray(final T[] a) {
		return getSortedList().toArray(a);
	}

	/* (non-Javadoc)
	 * @see java.util.List#add(java.lang.Object)
	 */
	@Override
	public boolean add(final E e) {
		DependencyIdentifier identifier = e.getIdentifier();
		this.dependancyMap.put(identifier, e);
		isSorted = false;
		return true;
	}

	/* (non-Javadoc)
	 * @see java.util.List#remove(java.lang.Object)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean remove(final Object o) {
		try {
			E dependency = (E) o;
			DependencyIdentifier identifier = dependency.getIdentifier();
			boolean removed = this.dependancyMap.remove(identifier) != null;
			if (removed) {
				isSorted = false;
			}
			return removed;
		} catch (ClassCastException e) {
			return false;
		}
	}

	/* (non-Javadoc)
	 * @see java.util.List#containsAll(java.util.Collection)
	 */
	@Override
	public boolean containsAll(final Collection<?> c) {
		return getSortedList().containsAll(c);
	}

	/* (non-Javadoc)
	 * @see java.util.List#addAll(java.util.Collection)
	 */
	@Override
	public boolean addAll(final Collection<? extends E> c) {
		for (E e : c) {
			add(e);
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see java.util.List#removeAll(java.util.Collection)
	 */
	@Override
	public boolean removeAll(final Collection<?> c) {
		boolean changed = false;
		for (Object object : c) {
			changed |= remove(object);
		}
		return changed;
	}

	/* (non-Javadoc)
	 * @see java.util.List#retainAll(java.util.Collection)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean retainAll(final Collection<?> c) {
		List<DependencyIdentifier> retainedIds = new ArrayList<DependencyIdentifier>();
		for (Object o : c) {
			try {
				E dependency = (E) o;
				DependencyIdentifier identifier = dependency.getIdentifier();
				retainedIds.add(identifier);
			} catch (ClassCastException e) {
			}
		}
		Set<DependencyIdentifier> keySet = dependancyMap.keySet();
		keySet.removeAll(retainedIds);
		for (DependencyIdentifier id : keySet) {
			dependancyMap.remove(id);
		}
		isSorted = keySet.isEmpty();
		return !keySet.isEmpty();
	}

	public E get(final int index) {
		return getSortedList().get(index);
	}

	public int indexOf(final Object o) {
		return getSortedList().indexOf(o);
	}

	/* (non-Javadoc)
	 * @see java.util.List#clear()
	 */
	@Override
	public void clear() {
		isSorted = false;
		dependancyMap.clear();
	}

	/* (non-Javadoc)
	 * @see java.util.SortedSet#comparator()
	 */
	@Override
	public Comparator<? super E> comparator() {
		return new Comparator<E>() {

			@Override
			public int compare(final E o1, final E o2) {
				if (dependsOn(o1, o2)) {
					return -1;
				} else if (dependsOn(o2, o1)) {
					return 1;
				}
				return 0;
			}

			private boolean dependsOn(final E o1, final E o2) {
				DependencyIdentifier identifier = o2.getIdentifier();
				List<? extends DependencyIdentifier> dependsOn = o1.dependsOn();
				if (dependsOn.isEmpty()) {
					return false;
				}
				if (dependsOn.contains(identifier)) {
					return true;
				}
				for (DependencyIdentifier dependencyIdentifier : dependsOn) {
					E e = dependancyMap.get(dependencyIdentifier);
					if (dependsOn(o1, e)) {
						return true;
					}
				}
				return false;
			}
		};
	}

	/* (non-Javadoc)
	 * @see java.util.SortedSet#subSet(java.lang.Object, java.lang.Object)
	 */
	@Override
	public SortedSet<E> subSet(final E fromElement, final E toElement) {
		List<E> sortedList = getSortedList();
		int fromIndex = sortedList.indexOf(fromElement);
		int toIndex = sortedList.indexOf(toElement);
		List<E> subList = sortedList.subList(fromIndex, toIndex);
		return new DependencySet<E>(subList);
	}

	/* (non-Javadoc)
	 * @see java.util.SortedSet#headSet(java.lang.Object)
	 */
	@Override
	public SortedSet<E> headSet(final E toElement) {
		List<E> sortedList = getSortedList();
		int toIndex = sortedList.indexOf(toElement);
		List<E> subList = sortedList.subList(0, toIndex);
		return new DependencySet<E>(subList);
	}

	/* (non-Javadoc)
	 * @see java.util.SortedSet#tailSet(java.lang.Object)
	 */
	@Override
	public SortedSet<E> tailSet(final E fromElement) {
		List<E> sortedList = getSortedList();
		int fromIndex = sortedList.indexOf(fromElement);
		List<E> subList = sortedList.subList(fromIndex, size() - 1);
		return new DependencySet<E>(subList);
	}

	/* (non-Javadoc)
	 * @see java.util.SortedSet#first()
	 */
	@Override
	public E first() {
		return getSortedList().get(0);
	}

	/* (non-Javadoc)
	 * @see java.util.SortedSet#last()
	 */
	@Override
	public E last() {
		return getSortedList().get(size() - 1);
	}

	private List<E> getSortedList() {
		if (!isSorted) {
			sort();
			isSorted = true;
		}
		return Collections.unmodifiableList(sorted);
	}

	private void sort() {
		sorted.clear();
		for (DependencyIdentifier identifier : dependancyMap.keySet()) {
			visit(identifier);
		}
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return getSortedList().toString();
	}

	private void visit(final DependencyIdentifier identifier) {
		if (!hasBeenVisited(identifier)) {
			markVisited(identifier);
			E d = dependancyMap.get(identifier);
			List<? extends DependencyIdentifier> dependsOn = d.dependsOn();
			for (DependencyIdentifier dependantIdentifier : dependsOn) {
				visit(dependantIdentifier);
			}
			sorted.add(d);
		}
	}

	private boolean hasBeenVisited(final DependencyIdentifier identifier) {
		return visited.contains(identifier);
	}

	private void markVisited(final DependencyIdentifier identifier) {
		this.visited.add(identifier);
	}
}
