/*
 * Copyright 2011 JOHK Software
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.wesleyhome.johksoftware.utilities.dependency;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The <code>TopologicalSorter</code> class will take a list of {@link Dependency} objects
 * and sort them based on their dependencies.
 * 
 * The {@link TopologicalSorter} implements {@link Iterable}
 * 
 * @author Justin Wesley
 * @since 1.0.0
 */
public class TopologicalSorter<E extends Dependency<I>, I extends DependencyIdentifier> implements Iterable<E> {

	private Map<I, E> dependancyMap;
	private List<I> visited;
	private List<E> sorted;
	private boolean isSorted = false;

	/**
	 * 
	 */
	public TopologicalSorter(final Set<Class<? extends E>> dependencyClassList) {
		super();
		List<E> dependancyList = new ArrayList<E>();
		for (Class<? extends E> class1 : dependencyClassList) {
			try {
				dependancyList.add(class1.newInstance());
			} catch (InstantiationException e) {
				throw new RuntimeException(e);
			} catch (IllegalAccessException e) {
				throw new RuntimeException(e);
			}
		}
		init(dependancyList);
	}

	/**
	 * @param dependancyList
	 */
	public TopologicalSorter(final List<E> dependancyList) {
		super();
		init(dependancyList);
	}

	/**
	 * @param dependancyList
	 */
	private void init(final List<E> dependancyList) {
		this.dependancyMap = new LinkedHashMap<I, E>();
		for (E e : dependancyList) {
			dependancyMap.put(e.getIdentifier(), e);
		}
		visited = new ArrayList<I>();
		sorted = new ArrayList<E>();
	}

	private void sort() {
		for (I identifier : dependancyMap.keySet()) {
			visit(identifier);
		}
	}

	private void visit(final I identifier) {
		if (!hasBeenVisited(identifier)) {
			markVisited(identifier);
			E d = dependancyMap.get(identifier);
			List<I> dependsOn = d.dependsOn();
			for (I dependantIdentifier : dependsOn) {
				visit(dependantIdentifier);
			}
			sorted.add(d);
		}
	}

	public List<E> getList() {
		if (!isSorted) {
			sort();
			isSorted = true;
		}
		return Collections.unmodifiableList(sorted);
	}

	public int size() {
		return getList().size();
	}

	/* (non-Javadoc)
	 * @see java.lang.Iterable#iterator()
	 */
	@Override
	public Iterator<E> iterator() {
		return getList().iterator();
	}

	private boolean hasBeenVisited(final I identifier) {
		return visited.contains(identifier);
	}

	private void markVisited(final I identifier) {
		this.visited.add(identifier);
	}
}
