/*
 * Copyright 2011 JOHK Software
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.wesleyhome.johksoftware.utilities.reflection;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.reflect.FieldUtils;
import sun.reflect.generics.reflectiveObjects.ParameterizedTypeImpl;


@SuppressWarnings("restriction")
public final class ReflectionUtil {

	public static Class<?> getPropertyType(final Class<?> masterClass, final String property) throws IntrospectionException {
		Field field = FieldUtils.getField(masterClass, property, true);
		if (field == null) {
			return getReturnType(getReadMethod(masterClass, property), masterClass);
		}
		Type genericType = field.getGenericType();
		if (genericType instanceof ParameterizedType) {
			ParameterizedType parameterizedType = (ParameterizedType) genericType;
			genericType = parameterizedType.getActualTypeArguments()[0];
		}
		if (!(genericType instanceof TypeVariable)) {
			return field.getType();
		}
		TypeVariable<?> type = (TypeVariable<?>) genericType;
		return getType(type, masterClass);
	}

	public static Class<?> getReturnType(final String readMethodName, final Class<?> masterClass) {
		return getReturnType(getMethod(readMethodName, masterClass), masterClass);
	}

	private static Class<?> getType(final TypeVariable<?> type, final Class<?> masterClass) {
		String variableName = type.getName();
		TypeVariable<?>[] typeParameters = masterClass.getSuperclass().getTypeParameters();
		Integer typeVariableIndex = null;
		for (int i = 0; i < typeParameters.length; i++) {
			TypeVariable<?> typeVariable = typeParameters[i];
			String typeName = typeVariable.getName();
			if (variableName.equals(typeName)) {
				typeVariableIndex = i;
				break;
			}
		}
		ParameterizedTypeImpl genericSuperclass = (ParameterizedTypeImpl) masterClass.getGenericSuperclass();
		Type[] actualTypeArguments = genericSuperclass.getActualTypeArguments();
		Type actualType = actualTypeArguments[typeVariableIndex];
		return (Class<?>) actualType;
	}

	public static Class<?> getReturnType(final Method readMethod, final Class<?> masterClass) {
		Type returnType = readMethod.getGenericReturnType();
		if (returnType instanceof ParameterizedType) {
			ParameterizedType parameterizedType = (ParameterizedType) returnType;
			returnType = parameterizedType.getActualTypeArguments()[0];
		}
		if (!(returnType instanceof TypeVariable)) {
			return readMethod.getReturnType();
		}
		TypeVariable<?> newReturnType = (TypeVariable<?>) returnType;
		return getType(newReturnType, masterClass);
	}

	public static Map<Class<? extends Annotation>, Annotation> getMethodAnnotations(final String methodName, final Class<?> cls) {
		List<Method> methods = getAllMethods(methodName, cls);
		Map<Class<? extends Annotation>, Annotation> map = new HashMap<Class<? extends Annotation>, Annotation>();
		for (Method method : methods) {
			String name = method.getName();
			if (methodName.equals(name)) {
				List<Annotation> annotationList = new ArrayList<Annotation>();
				annotationList.addAll(Arrays.asList(method.getAnnotations()));
				annotationList.addAll(Arrays.asList(method.getDeclaredAnnotations()));
				for (Annotation annotation : annotationList) {
					map.put(annotation.annotationType(), annotation);
				}
			}
		}
		return map;
	}

	public static List<Field> getAllFields(final Class<?> cls) {
		List<Field> fields = new ArrayList<Field>();
		fields.addAll(Arrays.asList(cls.getDeclaredFields()));
		Class<?> superclass = cls.getSuperclass();
		if (superclass != null) {
			fields.addAll(getAllFields(superclass));
		}
		return fields;
	}

	public static List<Method> getAllMethods(final String name, final Class<?> cls) {
		Class<?> superclass = cls.getSuperclass();
		List<Method> methods = new ArrayList<Method>();
		if (name == null) {
			return methods;
		}
		if (superclass != null) {
			methods.addAll(getAllMethods(name, superclass));
		}
		methods.addAll(Arrays.asList(cls.getMethods()));
		methods.addAll(Arrays.asList(cls.getDeclaredMethods()));
		Iterator<Method> itr = methods.iterator();
		while (itr.hasNext()) {
			String methodName = itr.next().getName();
			if (!name.equals(methodName)) {
				itr.remove();
			}
		}
		return methods;
	}

	public static Method getMethod(final String method, final Class<?> cls) {
		return getMethod(method, cls, new MethodPredicateAdapter());
	}

	public static Method getMethod(final String method, final Class<?> cls, final MethodPredicate predicate) {
		List<Method> allMethods = getAllMethods(method, cls);
		for (Method m : allMethods) {
			if (predicate.evaluate(m)) {
				return m;
			}
		}
		throw new IllegalArgumentException(String.format("%s method is not available of %s", method, cls.getSimpleName()));
	}

	public static Method getReadMethod(final Class<?> cls, final String property) throws IntrospectionException {
		return getReadMethod(cls, property, new MethodPredicateAdapter());
	}

	@SuppressWarnings("unchecked")
	public static Method getReadMethod(final Class<?> cls, final String property, final MethodPredicate predicate)
		throws IntrospectionException {
		PropertyDescriptor pd = new PropertyDescriptor(property, cls);
		final Method readMethod = pd.getReadMethod();
		List<Method> methods = Arrays.asList(ArrayUtils.addAll(cls.getDeclaredMethods(), cls.getMethods()));
		methods = (List<Method>) CollectionUtils.select(methods, new Predicate() {

			@Override
			public boolean evaluate(final Object o) {
				return ((Method) o).getName().equals(readMethod.getName());
			}
		});
		if (predicate != null) {
			for (Method method : methods) {
				if (predicate.evaluate(method)) {
					return method;
				}
			}
		}
		return methods.get(0);
	}

	public static List<Class<?>> getClassGenericTypes(final Class<?> cls) {
		List<Class<?>> classes = new ArrayList<Class<?>>();
		if (cls.isInterface()) {
			Type[] superInterfaces = cls.getGenericInterfaces();
			for (Type type : superInterfaces) {
				populateClasses(classes, type);
			}
		} else {
			Type superClass = cls.getGenericSuperclass();
			populateClasses(classes, superClass);
			Type[] superInterfaces = cls.getGenericInterfaces();
			for (Type type : superInterfaces) {
				populateClasses(classes, type);
			}
		}
		return classes;
	}

	/**
	 * @param classes
	 * @param superClass
	 */
	private static void populateClasses(final List<Class<?>> classes, final Type superClass) {
		if (superClass instanceof Class) {
			return;
		}
		ParameterizedTypeImpl genericSuperclass = (ParameterizedTypeImpl) superClass;
		Type[] types = genericSuperclass.getActualTypeArguments();
		for (Type type : types) {
			if (type instanceof Class) {
				classes.add((Class<?>) type);
			}
		}
	}

	/**
	 * @param propertyType
	 * @param class1
	 * @return
	 * @throws IntrospectionException 
	 */
	public static String getAnnotatedMethodProperty(final Class<?> propertyType, final Class<? extends Annotation> class1)
		throws IntrospectionException {
		PropertyDescriptor[] propertyDescriptors = Introspector.getBeanInfo(propertyType).getPropertyDescriptors();
		for (PropertyDescriptor pd : propertyDescriptors) {
			if (pd.getReadMethod().isAnnotationPresent(class1)) {
				return pd.getName();
			}
		}
		return null;
	}

	public static List<Field> getAnnotatedFields(final Class<?> annotatedClass, final Class<? extends Annotation> annotationClass) {
		List<Field> fields = new ArrayList<Field>();
		List<Field> allFields = getAllFields(annotatedClass);
		for (Field field : allFields) {
			if (field.isAnnotationPresent(annotationClass)) {
				fields.add(field);
			}
		}
		return fields;
	}
}
